
local addonName = ...

-- Default settings
local DEFAULTS = {
	minDuration = 1.5,
	fadedAlpha = 0.15,
	effectDuration = 5,
}

-- Current settings
local prefs = {}

-- Globals made upvalues
local _G = _G
local GetTime = GetTime
local next = next
local pairs = pairs

local running = {}
local frame = CreateFrame("Frame")
frame:Hide()

local pow = math.pow
local function Cooldown_OnUpdate(cooldown, now)
	local timeLeft = (cooldown.CF_expires or 0) - now
	local duration = min(prefs.effectDuration, cooldown.CF_duration)
	local alpha = cooldown.CF_alpha
	if timeLeft > duration then
		alpha = alpha * prefs.fadedAlpha
	elseif timeLeft > 0 then
		local t = pow(1.0 - timeLeft / duration, 2)
		alpha = alpha * prefs.fadedAlpha + t * (1.0 - prefs.fadedAlpha)
	end
	cooldown:GetParent():SetAlpha(alpha)
	if timeLeft <= 0 then
		cooldown:Hide()
	end
end

local function Cooldown_OnShow(cooldown)
	if (cooldown.CF_expires or 0) > GetTime() and not running[cooldown] then
		running[cooldown] = true
		cooldown.CF_alpha = cooldown.CF_alpha or cooldown:GetParent():GetAlpha()
		frame:Show()
	end
end

local function Cooldown_OnHide(cooldown)
	if running[cooldown] then
		running[cooldown] = nil
		if cooldown.CF_alpha then
			cooldown:GetParent():SetAlpha(cooldown.CF_alpha)
			cooldown.CF_alpha = nil
		end
	end
end

local frames = setmetatable({}, { __index = function(t, cooldown)
	local frame = false
	if not cooldown.noCooldownCount and not cooldown.noCooldownFader then
		frame = cooldown:GetParent()
		local name = frame:GetName()
		if name and (strmatch(name, '[bB]uff%d*$') or strmatch(name, '[bB]uffIcon%d*$')) then
			frame = false
		else
			cooldown:HookScript('OnShow', Cooldown_OnShow)
			cooldown:HookScript('OnHide', Cooldown_OnHide)
		end
	end
	t[cooldown] = frame
	return frame
end})

local function updateCooldown(cooldown, start, duration)
	local frame = frames[cooldown]
	if not frame then
		return -- do not mess with auras
	end
	
	if start < 0 or duration <= prefs.minDuration or start + duration < GetTime() then
		start, duration = nil, nil
	end
	
	local expires = start and duration and start + duration
	
	if expires ~= cooldown.CF_expires or duration ~= cooldown.CF_duration then
		cooldown.CF_expires, cooldown.CF_duration = expires, duration
		Cooldown_OnShow(cooldown)
	end
end

-- OnUpdate frames
local delay = 0
frame:SetScript('OnShow', function() delay = 0 end)
frame:SetScript('OnUpdate', function(self, elapsed)
	delay = delay - elapsed
	if delay > 0 then
		return
	end
	local now = GetTime()
	for cooldown in pairs(running) do
		Cooldown_OnUpdate(cooldown, now)
	end
	if not next(running) then
		self:Hide()
	else
		delay = 0.1
	end
end)

-- Initialization and setting handling

frame:RegisterEvent('ADDON_LOADED')
frame:SetScript('OnEvent', function(self, event, arg1)
	if event == 'ADDON_LOADED' and arg1 == addonName then
		self:UnregisterEvent('ADDON_LOADED')
		self:RegisterEvent('PLAYER_LOGOUT')

		-- Load settings and apply defaults
		for k, v in pairs(DEFAULTS) do
			prefs[k] = v
		end
		if _G.CooldownFaderDB then
			for k, v in pairs(_G.CooldownFaderDB) do
				prefs[k] = v
			end
		end

		-- Create a chat command
		_G.SLASH_COOLDOWNFADER1 = "/cooldownfader"
		_G.SLASH_COOLDOWNFADER2 = "/cdf"
		_G.SlashCmdList.COOLDOWNFADER = function(arg)
			local key, value = strsplit(" ", arg or "")
			if prefs[key] then
				value = tonumber(value)
				if value then
					prefs[key] = value
				end
				print(format('|cff33ff99CooldownFader:|r %s set to %g.', key, prefs[key]))
			else
				print('|cff33ff99CooldownFader settings:|r')
				for key, value in pairs(prefs) do
					print(format('  /cooldownfader %s [value] (current: %g, default: %g).', key, value, DEFAULTS[key]))
				end
			end
		end

		-- Put our hook in place
		hooksecurefunc(getmetatable(ActionButton1Cooldown).__index, 'SetCooldown', updateCooldown)

	elseif event == 'PLAYER_LOGOUT' then
		-- Recreate the saved variable
		_G.CooldownFaderDB = nil
		for k, v in pairs(prefs) do
			if v ~= DEFAULTS[k] then
				if _G.CooldownFaderDB then
					_G.CooldownFaderDB[k] = v
				else
					_G.CooldownFaderDB = { [k] = v }
				end
			end
		end
	end
end)
